@@ -1,4 +1,4 @@
-#     Copyright 2014, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2018, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -34,14 +34,32 @@ import sys
 def printIndented(level, *what):
     print("    " * level, *what)
 
+
 def printSeparator(level = 0):
-    print("    " * level, "*" * 10)
+    print("    " * level, '*' * 10)
+
 
 def printLine(*what):
     print(*what)
 
+
 def printError(message):
     print(
         message,
         file = sys.stderr
     )
+
+
+def flushStdout():
+    sys.stdout.flush()
+
+
+def my_print(*args, **kwargs):
+    """ Make sure we flush after every print.
+
+    Not even the "-u" option does more than that and this is easy enough.
+    """
+
+    print(*args, **kwargs)
+
+    flushStdout()
