@@ -1,4 +1,4 @@
-#     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2018, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -65,6 +65,14 @@ class ExpressionVariableNameRef(ExpressionBase):
 
         self.provider = provider
 
+    def finalize(self):
+        del self.parent
+        del self.provider
+
+    @staticmethod
+    def isExpressionVariableNameRef():
+        return True
+
     def getDetails(self):
         return {
             "variable_name" : self.variable_name,
@@ -77,6 +85,25 @@ class ExpressionVariableNameRef(ExpressionBase):
     def computeExpressionRaw(self, trace_collection):
         return self, None, None
 
+    @staticmethod
+    def needsFallback():
+        return True
+
+
+class ExpressionVariableLocalNameRef(ExpressionVariableNameRef):
+    """ These are used before the actual variable object is known from VariableClosure.
+
+        The special thing about this as opposed to ExpressionVariableNameRef is that
+        these must remain local names and cannot fallback to outside scopes. This is
+        used for __annotations__.
+
+    """
+    kind = "EXPRESSION_VARIABLE_LOCAL_NAME_REF"
+
+    @staticmethod
+    def needsFallback():
+        return False
+
 
 class ExpressionVariableRefBase(ExpressionBase):
     # Base classes can be abstract, pylint: disable=abstract-method
@@ -92,14 +119,134 @@ class ExpressionVariableRefBase(ExpressionBase):
         self.variable = variable
         self.variable_trace = None
 
+    def finalize(self):
+        del self.parent
+        del self.variable
+        del self.variable_trace
+
     def getVariableName(self):
         return self.variable.getName()
 
     def getVariable(self):
         return self.variable
 
-    def getVariableVersion(self):
-        return self.variable_trace.getVersion()
+    def getVariableTrace(self):
+        return self.variable_trace
+
+    def computeExpressionComparisonIn(self, in_node, value_node, trace_collection):
+        tags = None
+        message = None
+
+        # Any code could be run, note that.
+        trace_collection.onControlFlowEscape(in_node)
+
+        if self.variable_trace.hasShapeDictionaryExact():
+            tags = "new_expression"
+            message = """\
+Check '%s' on dictionary lowered to dictionary '%s'.""" % (
+                in_node.comparator,
+                in_node.comparator
+            )
+
+            if in_node.comparator == "In":
+                in_node = ExpressionDictOperationIn(
+                    key        = value_node,
+                    dict_arg   = self,
+                    source_ref = in_node.getSourceReference()
+                )
+            else:
+                in_node = ExpressionDictOperationNOTIn(
+                    key        = value_node,
+                    dict_arg   = self,
+                    source_ref = in_node.getSourceReference()
+                )
+
+
+        # Any exception may be raised.
+        if in_node.mayRaiseException(BaseException):
+            trace_collection.onExceptionRaiseExit(BaseException)
+
+        return in_node, tags, message
+
+    def computeExpressionSetSubscript(self, set_node, subscript, value_node,
+                                      trace_collection):
+        tags = None
+        message = None
+
+        # By default, an subscript may change everything about the lookup
+        # source.
+        if self.variable_trace.hasShapeDictionaryExact():
+            set_node = StatementDictOperationSet(
+                dict_arg   = self,
+                key        = subscript,
+                value      = value_node,
+                source_ref = set_node.getSourceReference()
+            )
+
+            tags = "new_statements"
+            message = """\
+Subscript assignment to dictionary lowered to dictionary assignment."""
+
+        # Any code could be run, note that.
+        trace_collection.onControlFlowEscape(self)
+
+        # Any exception might be raised.
+        if set_node.mayRaiseException(BaseException):
+            trace_collection.onExceptionRaiseExit(BaseException)
+
+        return set_node, tags, message
+
+    def computeExpressionDelSubscript(self, del_node, subscript,
+                                      trace_collection):
+        tags = None
+        message = None
+
+        # By default, an subscript may change everything about the lookup
+        # source.
+        # Any code could be run, note that.
+        trace_collection.onControlFlowEscape(self)
+
+        if self.variable_trace.hasShapeDictionaryExact():
+            del_node = StatementDictOperationRemove(
+                dict_arg   = self,
+                key        = subscript,
+                source_ref = del_node.getSourceReference()
+            )
+
+            tags = "new_statements"
+            message = """\
+Subscript del to dictionary lowered to dictionary del."""
+
+
+        # Any exception might be raised.
+        if del_node.mayRaiseException(BaseException):
+            trace_collection.onExceptionRaiseExit(BaseException)
+
+        return del_node, tags, message
+
+    def computeExpressionSubscript(self, lookup_node, subscript, trace_collection):
+        tags = None
+        message = None
+
+        # Any code could be run, note that.
+        trace_collection.onControlFlowEscape(self)
+
+        if self.variable_trace.hasShapeDictionaryExact():
+            lookup_node = ExpressionDictOperationGet(
+                dict_arg   = self,
+                key        = subscript,
+                source_ref = lookup_node.getSourceReference()
+            )
+
+            tags = "new_expression"
+            message = """\
+Subscript look-up to dictionary lowered to dictionary look-up."""
+
+        # Any exception might be raised.
+        if lookup_node.mayRaiseException(BaseException):
+            trace_collection.onExceptionRaiseExit(BaseException)
+
+        return lookup_node, tags, message
 
 
 class ExpressionVariableRef(ExpressionVariableRefBase):
@@ -228,7 +375,7 @@ Module variable '%s' found to be built-in reference.""" % (
                 )
             elif variable_name == "__name__":
                 new_node = ExpressionModuleAttributeNameRef(
-                    module     = variable.getOwner(),
+                    variable   = variable,
                     source_ref = self.getSourceReference()
                 )
 
@@ -237,16 +384,16 @@ Module variable '%s' found to be built-in reference.""" % (
 Replaced read-only module attribute '__name__' with module attribute reference."""
             elif variable_name == "__package__":
                 new_node = ExpressionModuleAttributePackageRef(
-                    module     = variable.getOwner(),
+                    variable   = variable,
                     source_ref = self.getSourceReference()
                 )
 
                 change_tags = "new_expression"
                 change_desc = """\
 Replaced read-only module attribute '__package__' with module attribute reference."""
-            elif variable_name == "__loader__" and python_version >= 330:
+            elif variable_name == "__loader__" and python_version >= 300:
                 new_node = ExpressionModuleAttributeLoaderRef(
-                    module     = variable.getOwner(),
+                    variable   = variable,
                     source_ref = self.getSourceReference()
                 )
 
@@ -255,7 +402,7 @@ Replaced read-only module attribute '__package__' with module attribute referenc
 Replaced read-only module attribute '__loader__' with module attribute reference."""
             elif variable_name == "__spec__" and python_version >= 340:
                 new_node = ExpressionModuleAttributeSpecRef(
-                    module     = variable.getOwner(),
+                    variable   = variable,
                     source_ref = self.getSourceReference()
                 )
 
@@ -305,121 +452,6 @@ local variable '%s' referenced before assignment""" % variable_name
 
         return call_node, None, None
 
-    def computeExpressionSetSubscript(self, set_node, subscript, value_node,
-                                      trace_collection):
-        tags = None
-        message = None
-
-        # By default, an subscript may change everything about the lookup
-        # source.
-        if self.variable_trace.hasShapeDictionaryExact():
-            set_node = StatementDictOperationSet(
-                dict_arg   = self,
-                key        = subscript,
-                value      = value_node,
-                source_ref = set_node.getSourceReference()
-            )
-
-            tags = "new_statements"
-            message = """\
-Subscript assignment to dictionary lowered to dictionary assignment."""
-
-        # Any code could be run, note that.
-        trace_collection.onControlFlowEscape(self)
-
-        # Any exception might be raised.
-        if set_node.mayRaiseException(BaseException):
-            trace_collection.onExceptionRaiseExit(BaseException)
-
-        return set_node, tags, message
-
-    def computeExpressionDelSubscript(self, del_node, subscript,
-                                      trace_collection):
-        tags = None
-        message = None
-
-        # By default, an subscript may change everything about the lookup
-        # source.
-        # Any code could be run, note that.
-        trace_collection.onControlFlowEscape(self)
-
-        if self.variable_trace.hasShapeDictionaryExact():
-            del_node = StatementDictOperationRemove(
-                dict_arg   = self,
-                key        = subscript,
-                source_ref = del_node.getSourceReference()
-            )
-
-            tags = "new_statements"
-            message = """\
-Subscript del to dictionary lowered to dictionary del."""
-
-
-        # Any exception might be raised.
-        if del_node.mayRaiseException(BaseException):
-            trace_collection.onExceptionRaiseExit(BaseException)
-
-        return del_node, tags, message
-
-    def computeExpressionSubscript(self, lookup_node, subscript, trace_collection):
-        tags = None
-        message = None
-
-        # Any code could be run, note that.
-        trace_collection.onControlFlowEscape(self)
-
-        if self.variable_trace.hasShapeDictionaryExact():
-            lookup_node = ExpressionDictOperationGet(
-                dict_arg   = self,
-                key        = subscript,
-                source_ref = lookup_node.getSourceReference()
-            )
-
-            tags = "new_expression"
-            message = """\
-Subscript look-up to dictionary lowered to dictionary look-up."""
-
-        # Any exception might be raised.
-        if lookup_node.mayRaiseException(BaseException):
-            trace_collection.onExceptionRaiseExit(BaseException)
-
-        return lookup_node, tags, message
-
-    def computeExpressionComparisonIn(self, in_node, value_node, trace_collection):
-        tags = None
-        message = None
-
-        # Any code could be run, note that.
-        trace_collection.onControlFlowEscape(in_node)
-
-        if self.variable_trace.hasShapeDictionaryExact():
-            tags = "new_expression"
-            message = """\
-Check '%s' on dictionary lowered to dictionary '%s'.""" % (
-                in_node.comparator,
-                in_node.comparator
-            )
-
-            if in_node.comparator == "In":
-                in_node = ExpressionDictOperationIn(
-                    key        = value_node,
-                    dict_arg   = self,
-                    source_ref = in_node.getSourceReference()
-                )
-            else:
-                in_node = ExpressionDictOperationNOTIn(
-                    key        = value_node,
-                    dict_arg   = self,
-                    source_ref = in_node.getSourceReference()
-                )
-
-
-        # Any exception may be raised.
-        if in_node.mayRaiseException(BaseException):
-            trace_collection.onExceptionRaiseExit(BaseException)
-
-        return in_node, tags, message
-
     def hasShapeDictionaryExact(self):
         return self.variable_trace.hasShapeDictionaryExact()
 
@@ -433,9 +465,8 @@ Check '%s' on dictionary lowered to dictionary '%s'.""" % (
         return not self.variable_trace.mustHaveValue()
 
     def mayRaiseException(self, exception_type):
-        variable_trace = self.variable_trace
-
-        return variable_trace is None or not self.variable_trace.mustHaveValue()
+        return self.variable_trace is None or \
+               not self.variable_trace.mustHaveValue()
 
 
 class ExpressionTempVariableRef(ExpressionVariableRefBase):
